#include "theBiasPlanet/coreUtilities/collections/NavigableLinkedMap.hpp"
#include <cstdarg>
#include <stdexcept>
#include <string>

namespace theBiasPlanet {
	namespace coreUtilities {
		namespace collections {
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::Links::Links (): i_previousKey (nullptr), i_nextKey (nullptr) {
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::Links::Links (key_type const * const a_previousKey, key_type const * const a_nextKey): i_previousKey (a_previousKey), i_nextKey (a_nextKey) {
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::Links::Links (Links const & a_copiedObject): i_previousKey (a_copiedObject.i_previousKey), i_nextKey (a_copiedObject.i_nextKey) {
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::Links & NavigableLinkedMap <T, U, W>::Links::operator = (Links const & a_assignedFromObject) {
				i_previousKey = a_assignedFromObject.i_previousKey;
				i_nextKey = a_assignedFromObject.i_nextKey;
				return *this;
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::Links::~Links () {
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::key_type const * const NavigableLinkedMap <T, U, W>::Links::getPreviousKey () const {
				return i_previousKey;
			}
			
			template <typename T, typename U, typename W> void NavigableLinkedMap <T, U, W>::Links::setPreviousKey (key_type const * const a_previousKey) {
				i_previousKey = a_previousKey;
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::key_type const * const NavigableLinkedMap <T, U, W>::Links::getNextKey () const {
				return i_nextKey;
			}
			
			template <typename T, typename U, typename W> void NavigableLinkedMap <T, U, W>::Links::setNextKey (key_type const * const a_nextKey) {
				i_nextKey = a_nextKey;
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::KeyValuePairComparer::KeyValuePairComparer (key_compare const & a_keyComparer): i_keyComparer (a_keyComparer) {
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::KeyValuePairComparer::KeyValuePairComparer (KeyValuePairComparer const & a_copiedObject): i_keyComparer (a_copiedObject.i_keyComparer) {
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::KeyValuePairComparer & NavigableLinkedMap <T, U, W>::KeyValuePairComparer::operator = (KeyValuePairComparer const & a_assignedFromObject) {
				i_keyComparer = a_assignedFromObject.i_keyComparer;
				return *this;
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::KeyValuePairComparer::~KeyValuePairComparer () {
			}
			
			template <typename T, typename U, typename W> bool NavigableLinkedMap <T, U, W>::KeyValuePairComparer::operator () (value_type const & a_keyValuePair1, value_type const & a_keyValuePair2) const {
				return i_keyComparer (a_keyValuePair1.first, a_keyValuePair2.first);
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::BaseIterator::BaseIterator (NavigableLinkedMap <key_type, mapped_type, key_compare> const & a_navigableLinkedMap, key_type const * const a_currentKey): i_keyToValueMap (a_navigableLinkedMap.i_keyToValueMap), i_keyToLinksMap (a_navigableLinkedMap.i_keyToLinksMap), i_firstKey (a_navigableLinkedMap.i_firstKey), i_lastKey (a_navigableLinkedMap.i_lastKey), i_currentKey (a_currentKey) {
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::BaseIterator::BaseIterator (BaseIterator const & a_copiedObject): i_keyToValueMap (a_copiedObject.i_keyToValueMap), i_keyToLinksMap (a_copiedObject.i_keyToLinksMap), i_firstKey (a_copiedObject.i_firstKey), i_lastKey (a_copiedObject.i_lastKey), i_currentKey (a_copiedObject.i_currentKey) {
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::BaseIterator & NavigableLinkedMap <T, U, W>::BaseIterator::operator = (BaseIterator const & a_assignedFromObject) {
				i_keyToValueMap = a_assignedFromObject.i_keyToValueMap;
				i_keyToLinksMap = a_assignedFromObject.i_keyToLinksMap;
				i_firstKey = a_assignedFromObject.i_firstKey;
				i_lastKey = a_assignedFromObject.i_lastKey;
				i_currentKey = a_assignedFromObject.i_lastKey;
				return *this;
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::BaseIterator::~BaseIterator () {
			}
			
			template <typename T, typename U, typename W> void NavigableLinkedMap <T, U, W>::BaseIterator::goToPreviousElement () {
				if (i_currentKey != nullptr) {
					try {
						Links const & l_currentElementLinks = i_keyToLinksMap->at (i_currentKey);
						// If the current element is the first element, the past-the-end-element becomes the current element.
						i_currentKey = l_currentElementLinks.getPreviousKey ();
					}
					catch (out_of_range & l_exception) {
						i_currentKey  = nullptr;
					}
				}
				else {
					i_currentKey = i_lastKey;
				}
			}
			
			template <typename T, typename U, typename W> void NavigableLinkedMap <T, U, W>::BaseIterator::goToNextElement () {
				if (i_currentKey != nullptr) {
					try {
						Links const & l_currentElementLinks = i_keyToLinksMap->at (i_currentKey);
						i_currentKey = l_currentElementLinks.getNextKey ();
					}
					catch (out_of_range & l_exception) {
						i_currentKey  = nullptr;
					}
				}
				else {
				}
			}
			
			template <typename T, typename U, typename W> bool NavigableLinkedMap <T, U, W>::BaseIterator::operator == (BaseIterator const & a_iteratorToBeCompared) const {
				return (i_currentKey == a_iteratorToBeCompared.i_currentKey);
			}
			
			template <typename T, typename U, typename W> bool NavigableLinkedMap <T, U, W>::BaseIterator::operator != (BaseIterator const & a_iteratorToBeCompared) const {
				return (i_currentKey != a_iteratorToBeCompared.i_currentKey);
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::NonConstantIterator::NonConstantIterator (NavigableLinkedMap <key_type, mapped_type, key_compare> const & a_navigableLinkedMap, key_type const * const a_currentKey): BaseIterator (a_navigableLinkedMap, a_currentKey) {
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::NonConstantIterator::NonConstantIterator (NonConstantIterator const & a_copiedObject): BaseIterator (a_copiedObject) {
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::NonConstantIterator & NavigableLinkedMap <T, U, W>::NonConstantIterator::operator = (NonConstantIterator const & a_assignedFromObject) {
				BaseIterator::operator = (a_assignedFromObject);
				return *this;
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::NonConstantIterator::~NonConstantIterator () {
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::NonConstantIterator & NavigableLinkedMap <T, U, W>::NonConstantIterator::operator ++ () {
				this->goToNextElement ();
				return *this;
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::NonConstantIterator NavigableLinkedMap <T, U, W>::NonConstantIterator::operator ++ (int const) {
				NonConstantIterator l_copiedIterator (*this);
				this->goToNextElement ();
				return l_copiedIterator;
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::NonConstantIterator & NavigableLinkedMap <T, U, W>::NonConstantIterator::operator -- () {
				this->goToPreviousElement ();
				return *this;
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::NonConstantIterator NavigableLinkedMap <T, U, W>::NonConstantIterator::operator -- (int const) {
				NonConstantIterator l_copiedIterator (*this);
				this->goToPreviousElement ();
				return l_copiedIterator;
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::NonConstantIterator::value_type & NavigableLinkedMap <T, U, W>::NonConstantIterator::operator * () const {
				if (this->i_currentKey != nullptr) {
					try {
						return const_cast <value_type &> (this->i_keyToValueMap->find (*(this->i_currentKey)).operator * ());
					}
					catch (out_of_range & l_exception) {
						throw l_exception;
					}
				}
				else {
					throw out_of_range ("The key does not exist.");
				}
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::NonConstantIterator::value_type * const NavigableLinkedMap <T, U, W>::NonConstantIterator::operator -> () const {
				if (this->i_currentKey != nullptr) {
					try {
						return const_cast <value_type *> (this->i_keyToValueMap->find (*(this->i_currentKey)).operator -> ());
						
					}
					catch (out_of_range & l_exception) {
						throw l_exception;
					}
				}
				else {
					throw out_of_range ("The key does not exist.");
				}
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::ConstantIterator::ConstantIterator (NavigableLinkedMap <key_type, mapped_type, key_compare> const & a_navigableLinkedMap, key_type const * const a_currentKey): BaseIterator (a_navigableLinkedMap, a_currentKey) {
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::ConstantIterator::ConstantIterator (ConstantIterator const & a_copiedObject): BaseIterator (a_copiedObject) {
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::ConstantIterator & NavigableLinkedMap <T, U, W>::ConstantIterator::operator = (ConstantIterator const & a_assignedFromObject) {
				BaseIterator::operator = (a_assignedFromObject);
				return *this;
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::ConstantIterator::~ConstantIterator () {
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::ConstantIterator & NavigableLinkedMap <T, U, W>::ConstantIterator::operator ++ () {
				this->goToNextElement ();
				return *this;
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::ConstantIterator NavigableLinkedMap <T, U, W>::ConstantIterator::operator ++ (int const) {
				ConstantIterator l_copiedIterator (*this);
				this->goToNextElement ();
				return l_copiedIterator;
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::ConstantIterator & NavigableLinkedMap <T, U, W>::ConstantIterator::operator -- () {
				this->goToPreviousElement ();
				return *this;
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::ConstantIterator NavigableLinkedMap <T, U, W>::ConstantIterator::operator -- (int const) {
				ConstantIterator l_copiedIterator (*this);
				this->goToPreviousElement ();
				return l_copiedIterator;
			}
			
#ifdef GCC
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::ConstantIterator::value_type & NavigableLinkedMap <T, U, W>::ConstantIterator::operator * () const {
#else
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::ConstantIterator::NonConstantvalue_type const & NavigableLinkedMap <T, U, W>::ConstantIterator::operator * () const {
#endif
				if (this->i_currentKey != nullptr) {
					try {
						return this->i_keyToValueMap->find (*(this->i_currentKey)).operator * ();
					}
					catch (out_of_range & l_exception) {
						throw l_exception;
					}
				}
				else {
					throw out_of_range ("The key does not exist.");
				}
			}
			
#ifdef GCC
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::ConstantIterator::value_type * const NavigableLinkedMap <T, U, W>::ConstantIterator::operator -> () const {
#else
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::ConstantIterator::NonConstantvalue_type const * const NavigableLinkedMap <T, U, W>::ConstantIterator::operator -> () const {
#endif
				if (this->i_currentKey != nullptr) {
					try {
						return this->i_keyToValueMap->find (*(this->i_currentKey)).operator -> ();
						
					}
					catch (out_of_range & l_exception) {
						throw l_exception;
					}
				}
				else {
					throw out_of_range ("The key does not exist.");
				}
			}
			
			template <typename T, typename U, typename W> pair <typename NavigableLinkedMap <T, U, W>::key_type const &, typename NavigableLinkedMap <T, U, W>::mapped_type const &> NavigableLinkedMap <T, U, W>::getKeyValuePair (key_type const & a_key, mapped_type const & a_value) const {
				return pair <key_type const &, mapped_type const &> (a_key, a_value);
			}
			
			template <typename T, typename U, typename W> void NavigableLinkedMap <T, U, W>::insertNewElementAtEnd (key_type const & a_key, optional <mapped_type> const & a_value) {
				mapped_type & l_insertedElementValue = (*i_keyToValueMap) [a_key];
				key_type const & l_key = i_keyToValueMap->find (a_key)->first; // the stored key, not any copy of it
				Links & l_insertedElementLinks = (*i_keyToLinksMap) [&l_key];
				if (i_lastKey != nullptr) {
					(*i_keyToLinksMap) [i_lastKey].setNextKey (&l_key);
				}
				l_insertedElementLinks.setPreviousKey (i_lastKey); 
				i_lastKey = &l_key;
				if (i_firstKey == nullptr) {
					i_firstKey = &l_key;
				}
				if (a_value.has_value ()) {
					l_insertedElementValue = *a_value;	
				}
			}
			
			template <typename T, typename U, typename W> void NavigableLinkedMap <T, U, W>::insertNewElementAtMiddle (key_type const & a_positionKey, key_type const & a_key, optional <mapped_type> const & a_value) {
				key_type const & l_positionKey = i_keyToValueMap->find (a_positionKey)->first; // the stored key, not any copy of it
				Links & l_foundElementLinks = (*i_keyToLinksMap) [&l_positionKey];
				mapped_type & l_insertedElementValue = (*i_keyToValueMap) [a_key];
				key_type const & l_key = i_keyToValueMap->find (a_key)->first; // the stored key, not any copy of it
				Links & l_insertedElementLinks = (*i_keyToLinksMap) [&l_key];
				key_type const * const l_previousKey = l_foundElementLinks.getPreviousKey ();
				if (l_previousKey == nullptr) {
					i_firstKey = &l_key;
				}
				else {
					Links & l_previousElementLinks = (*i_keyToLinksMap) [l_previousKey];
					l_previousElementLinks.setNextKey (&l_key);
				}
				l_insertedElementLinks.setPreviousKey (l_foundElementLinks.getPreviousKey ()); 
				l_insertedElementLinks.setNextKey (&l_positionKey); 
				if (a_value.has_value ()) {
					l_insertedElementValue = *a_value; 
				}
				l_foundElementLinks.setPreviousKey (&l_key); 
			}
			
			template <typename T, typename U, typename W> void NavigableLinkedMap <T, U, W>::updateElement (key_type const & a_key, mapped_type const & a_value) {
				mapped_type & l_updatedElementValue = i_keyToValueMap->at (a_key);
				l_updatedElementValue = a_value; 
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::NavigableLinkedMap (key_compare const & a_keysComparer): i_keyToValueMap (new map <key_type, mapped_type, key_compare> (a_keysComparer)), i_keyToLinksMap (new map <key_type const * const, Links> ()), i_firstKey (nullptr), i_lastKey (nullptr) {
			}
			
			template <typename T, typename U, typename W> template <typename V> NavigableLinkedMap <T, U, W>::NavigableLinkedMap (V const & a_iteratorPointedAtFirstElement, V const & a_iteratorPointedAtLastElementNotIncluded, key_compare const & a_keysComparer): i_keyToValueMap (new map <key_type, mapped_type, key_compare> (a_keysComparer)), i_keyToLinksMap (new map <key_type const * const, Links> ()), i_firstKey (nullptr), i_lastKey (nullptr) {
				this->insert (a_iteratorPointedAtFirstElement, a_iteratorPointedAtLastElementNotIncluded);
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::NavigableLinkedMap (initializer_list <value_type> const & a_initializer, key_compare const & a_keysComparer): i_keyToValueMap (new map <key_type, mapped_type, key_compare> (a_keysComparer)), i_keyToLinksMap (new map <key_type const * const, Links> ()), i_firstKey (nullptr), i_lastKey (nullptr) {
				for (value_type const & l_keyValuePair: a_initializer) {
					insert (l_keyValuePair);
				}
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::NavigableLinkedMap (NavigableLinkedMap <key_type, mapped_type, key_compare> const & a_copiedObject): i_keyToValueMap (new map <key_type, mapped_type, key_compare> (a_copiedObject.key_comp ())), i_keyToLinksMap (new map <key_type const * const, Links> ()), i_firstKey (nullptr), i_lastKey (nullptr) {
				for (NavigableLinkedMap <key_type, mapped_type, key_compare>::value_type const & l_keyValuePair: a_copiedObject) {
					insert (l_keyValuePair);
				}
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W> & NavigableLinkedMap <T, U, W>::operator = (NavigableLinkedMap const & a_assignedFromObject) {
				if (i_keyToValueMap != nullptr) {
					delete i_keyToValueMap;
				}
				if (i_keyToLinksMap != nullptr) {
					delete i_keyToLinksMap;
				}
				i_keyToValueMap = new map <key_type, mapped_type, key_compare> (a_assignedFromObject.key_comp ());
				i_keyToLinksMap = new map <key_type const * const, Links> ();
				i_firstKey = nullptr;
				i_lastKey = nullptr;
				for (NavigableLinkedMap <key_type, mapped_type, key_compare>::value_type const & l_keyValuePair: a_assignedFromObject) {
					insert (l_keyValuePair);
				}
				return *this;
			}
			
			template <typename T, typename U, typename W> NavigableLinkedMap <T, U, W>::~NavigableLinkedMap () {
				if (i_keyToValueMap != nullptr) {
					delete i_keyToValueMap;
					i_keyToValueMap = nullptr;
				}
				if (i_keyToLinksMap != nullptr) {
					delete i_keyToLinksMap;
					i_keyToLinksMap = nullptr;
				}
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::iterator NavigableLinkedMap <T, U, W>::begin () {
				return iterator (*this, i_firstKey);
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::const_iterator NavigableLinkedMap <T, U, W>::begin () const {
				return const_iterator (*this, i_firstKey);
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::iterator NavigableLinkedMap <T, U, W>::end () {
				return iterator (*this, nullptr);
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::const_iterator NavigableLinkedMap <T, U, W>::end () const {
				return const_iterator (*this, nullptr);
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::reverse_iterator NavigableLinkedMap <T, U, W>::rbegin () {
				return reverse_iterator (end ());
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::const_reverse_iterator NavigableLinkedMap <T, U, W>::rbegin () const {
				return const_reverse_iterator (end ());
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::reverse_iterator NavigableLinkedMap <T, U, W>::rend () {
				return reverse_iterator (begin ());
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::const_reverse_iterator NavigableLinkedMap <T, U, W>::rend () const {
				return const_reverse_iterator (begin ());
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::const_iterator NavigableLinkedMap <T, U, W>::cbegin () const noexcept {
				return const_iterator (*this, i_firstKey);
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::const_iterator NavigableLinkedMap <T, U, W>::cend () const noexcept {
				return const_iterator (*this, nullptr);
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::const_reverse_iterator NavigableLinkedMap <T, U, W>::crbegin () const noexcept {
				return const_reverse_iterator (cend ());
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::const_reverse_iterator NavigableLinkedMap <T, U, W>::crend () const noexcept {
				return const_reverse_iterator (cbegin ());
			}
			
			template <typename T, typename U, typename W> bool NavigableLinkedMap <T, U, W>::empty () const {
				return i_keyToValueMap->empty ();
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::size_type NavigableLinkedMap <T, U, W>::size () const {
				return i_keyToValueMap->size ();
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::size_type NavigableLinkedMap <T, U, W>::max_size () const {
				return i_keyToValueMap->max_size ();
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::mapped_type & NavigableLinkedMap <T, U, W>::operator []  (key_type const & a_key) {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				typename map <key_type, mapped_type, key_compare>::iterator l_keyToValueMapIteratorAtKey = i_keyToValueMap->find (a_key);
				if (l_keyToValueMapIteratorAtKey == i_keyToValueMap->end ()) {
					insertNewElementAtEnd (a_key, nullopt);
					l_keyToValueMapIteratorAtKey = i_keyToValueMap->find (a_key);
				}
				return (*i_keyToValueMap) [l_keyToValueMapIteratorAtKey->first];
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::mapped_type & NavigableLinkedMap <T, U, W>::at (key_type const & a_key) {
				return i_keyToValueMap->at (a_key);
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::mapped_type const & NavigableLinkedMap <T, U, W>::at (key_type const & a_key) const {
				return i_keyToValueMap->at (a_key);
			}
			
			template <typename T, typename U, typename W> pair <typename NavigableLinkedMap <T, U, W>::iterator, bool> NavigableLinkedMap <T, U, W>::insert (value_type const & a_keyValuePair) {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				bool l_isNewlyInserted = false;
				typename map <key_type, mapped_type, key_compare>::iterator l_keyToValueMapIteratorAtKey = i_keyToValueMap->find (a_keyValuePair.first);
				if (l_keyToValueMapIteratorAtKey == i_keyToValueMap->end ()) {
					l_isNewlyInserted = true;
					insertNewElementAtEnd (a_keyValuePair.first, a_keyValuePair.second);
					l_keyToValueMapIteratorAtKey = i_keyToValueMap->find (a_keyValuePair.first);
				}
				else {
					updateElement (l_keyToValueMapIteratorAtKey->first, a_keyValuePair.second);
				}
				return pair <typename NavigableLinkedMap <T, U, W>::iterator, bool> (iterator (*this, & (l_keyToValueMapIteratorAtKey->first)), l_isNewlyInserted);
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::iterator NavigableLinkedMap <T, U, W>::insert (iterator a_position, value_type const & a_keyValuePair) {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				typename map <key_type, mapped_type, key_compare>::iterator l_keyToValueMapIteratorAtKey = i_keyToValueMap->find (a_keyValuePair.first);
				if (l_keyToValueMapIteratorAtKey == i_keyToValueMap->end ()) {
					typename map <key_type, mapped_type, key_compare>::iterator l_keyToValueMapIteratorAtPosition = i_keyToValueMap->find ( (*a_position).first);
					if (l_keyToValueMapIteratorAtPosition == i_keyToValueMap->end ()) {
						insertNewElementAtEnd (a_keyValuePair.first, a_keyValuePair.second);
					}
					else {
						insertNewElementAtMiddle ( (*a_position).first, a_keyValuePair.first, a_keyValuePair.second);
					}
				}
				else {
					updateElement (a_keyValuePair.first, a_keyValuePair.second);
				}
				l_keyToValueMapIteratorAtKey = i_keyToValueMap->find (a_keyValuePair.first);
				return iterator (*this, & (l_keyToValueMapIteratorAtKey->first));
			}
			
			template <typename T, typename U, typename W> template <typename V> void NavigableLinkedMap <T, U, W>::insert (V const & a_iteratorPointedAtFirstElement, V const & a_iteratorPointedAtLastElementNotIncluded) {
				for (V l_iterator = a_iteratorPointedAtFirstElement; l_iterator != a_iteratorPointedAtLastElementNotIncluded; l_iterator ++) {
					insert (*l_iterator);
				}
			}
			
			template <typename T, typename U, typename W> void NavigableLinkedMap <T, U, W>::erase (iterator a_position) {
				erase ((*a_position).first);
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::size_type NavigableLinkedMap <T, U, W>::erase (key_type const & a_key) {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				typename map <key_type, mapped_type, key_compare>::iterator l_keyToValueMapIteratorAtKey = i_keyToValueMap->find (a_key);
				if (l_keyToValueMapIteratorAtKey == i_keyToValueMap->end ()) {
					return 0;
				}
				else {
					key_type const & l_key = l_keyToValueMapIteratorAtKey->first; // the stored key, not any copy of it
					Links & l_foundElementLinks = (*i_keyToLinksMap) [&l_key];
					key_type const * const l_previousKey = l_foundElementLinks.getPreviousKey (); 
					key_type const * const l_nextKey = l_foundElementLinks.getNextKey (); 
					if (l_previousKey != nullptr) {
						Links & l_previousElementLinks = (*i_keyToLinksMap) [l_previousKey];
						if (l_nextKey != nullptr) {
							Links & l_nextElementLinks = (*i_keyToLinksMap) [l_nextKey];
							l_previousElementLinks.setNextKey (l_nextKey);
							l_nextElementLinks.setPreviousKey (l_previousKey);
						}
						else {
							l_previousElementLinks.setNextKey (nullptr);
							i_lastKey = l_previousKey;
						}
					}
					else {
						if (l_nextKey != nullptr) {
							Links & l_nextElementLinks = (*i_keyToLinksMap) [l_nextKey];
							i_firstKey = l_nextKey;
							l_nextElementLinks.setPreviousKey (nullptr);
						}
						else {
							i_firstKey = nullptr;
							i_lastKey = nullptr;
						}
					}
					i_keyToValueMap->erase (l_key);
					i_keyToLinksMap->erase (&l_key);
					return 1;
				}
			}
			
     		template <typename T, typename U, typename W> void NavigableLinkedMap <T, U, W>::erase (iterator a_iteratorPointedAtFirstElement, iterator a_iteratorPointedAtLastElementNotIncluded) {
				for (NavigableLinkedMap <key_type, mapped_type, key_compare>::iterator l_iterator = a_iteratorPointedAtFirstElement; l_iterator != a_iteratorPointedAtLastElementNotIncluded; l_iterator++) {
					erase (l_iterator);
				}
			}
			
			template <typename T, typename U, typename W> void NavigableLinkedMap <T, U, W>::swap (NavigableLinkedMap <key_type, mapped_type, key_compare> & a_navigableLinkedMap) {
				map <key_type, mapped_type, key_compare> * l_keyToValueMapSaved = i_keyToValueMap;
				map <key_type const * const, Links> * l_keyToLinksMapSaved = i_keyToLinksMap;
				i_keyToValueMap = a_navigableLinkedMap.i_keyToValueMap;
				i_keyToLinksMap = a_navigableLinkedMap.i_keyToLinksMap;
				a_navigableLinkedMap.i_keyToValueMap = l_keyToValueMapSaved;
				a_navigableLinkedMap.i_keyToLinksMap = l_keyToLinksMapSaved;
				key_type const * const l_firstKeySaved = i_firstKey;
				i_firstKey = a_navigableLinkedMap.i_firstKey;
				a_navigableLinkedMap.i_firstKey = l_firstKeySaved;
				key_type const * const l_lastKeySaved = i_lastKey;
				i_lastKey = a_navigableLinkedMap.i_lastKey;
				a_navigableLinkedMap.i_lastKey = l_lastKeySaved;
			}
			
			template <typename T, typename U, typename W> void NavigableLinkedMap <T, U, W>::clear () {
				i_keyToValueMap->clear ();
				i_keyToLinksMap->clear ();
				i_firstKey = nullptr;
				i_lastKey = nullptr;
			}
			
			template <typename T, typename U, typename W> template <typename ... V> pair <typename NavigableLinkedMap <T, U, W>::iterator, bool> NavigableLinkedMap <T, U, W>::emplace (V && ... a_arguments) {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				pair <key_type const &, mapped_type const &> l_keyValuePair = getKeyValuePair (a_arguments...);
				T const & l_keyToBeCopied = l_keyValuePair.first;
				U const & l_value = l_keyValuePair.second;
				bool l_isNewlyInserted = false;
				typename map <key_type, mapped_type, key_compare>::iterator l_keyToValueMapIteratorAtKeyToBeCopied = i_keyToValueMap->find (l_keyToBeCopied);
				if (l_keyToValueMapIteratorAtKeyToBeCopied == i_keyToValueMap->end ()) {
					l_isNewlyInserted = true;
					i_keyToValueMap->emplace (l_keyToBeCopied, l_value);
					key_type const & l_key = i_keyToValueMap->find (l_keyToBeCopied)->first; // the stored key, not any copy of it
					if (i_lastKey != nullptr) {
						(*i_keyToLinksMap) [i_lastKey].setNextKey (&l_key);
					}
					i_keyToLinksMap->emplace (&l_key, Links (i_lastKey, nullptr));
					i_lastKey = &l_key;
					if (i_firstKey == nullptr) {
						i_firstKey = &l_key;
					}
					return pair (iterator (*this, &l_key), l_isNewlyInserted);
				}
				else {
					return pair (iterator (*this, & (l_keyToValueMapIteratorAtKeyToBeCopied->first)), l_isNewlyInserted);
				}
			}
			
			template <typename T, typename U, typename W> template <typename ... V> typename NavigableLinkedMap <T, U, W>::iterator NavigableLinkedMap <T, U, W>::emplace_hint (const_iterator a_position, V && ... a_arguments) {
				unique_lock <recursive_mutex> l_lock (i_mutex);
				pair <key_type const &, mapped_type const &> l_keyValuePair = getKeyValuePair (a_arguments...);
				key_type const & l_keyToBeCopied = l_keyValuePair.first;
				mapped_type const & l_value = l_keyValuePair.second;
				typename map <key_type, mapped_type, key_compare>::iterator l_keyToValueMapIteratorAtKeyToBeCopied = i_keyToValueMap->find (l_keyToBeCopied);
				if (l_keyToValueMapIteratorAtKeyToBeCopied == i_keyToValueMap->end ()) {
					typename map <key_type, mapped_type, key_compare>::iterator l_keyToValueMapIteratorAtPosition = i_keyToValueMap->find (a_position->first);
					if (l_keyToValueMapIteratorAtPosition == i_keyToValueMap->end ()) {
						return emplace (a_arguments...).first;
					}
					else {
						key_type const & l_positionKey = l_keyToValueMapIteratorAtPosition->first; // the stored key, not any copy of it
						Links & l_foundElementLinks = (*i_keyToLinksMap) [&l_positionKey];
						key_type const * const l_previousKey = l_foundElementLinks.getPreviousKey ();
						i_keyToValueMap->emplace (l_keyToBeCopied, l_value);
						l_keyToValueMapIteratorAtKeyToBeCopied = i_keyToValueMap->find (l_keyToBeCopied);
						key_type const & l_key = l_keyToValueMapIteratorAtKeyToBeCopied->first; // the stored key, not any copy of it
						i_keyToLinksMap->emplace (&l_key, Links (l_previousKey, &l_positionKey));
						if (l_previousKey == nullptr) {
							i_firstKey = &l_key;
						}
						else {
							Links & l_previousElementLinks = (*i_keyToLinksMap) [l_previousKey];
							l_previousElementLinks.setNextKey (&l_key); 
						}
						l_foundElementLinks.setPreviousKey (&l_key); 
						return iterator (*this, &l_key);
					}
				}
				else {
					return iterator (*this, & (l_keyToValueMapIteratorAtKeyToBeCopied->first));
				}
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::key_compare NavigableLinkedMap <T, U, W>::key_comp () const {
				return i_keyToValueMap->key_comp ();
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::value_compare NavigableLinkedMap <T, U, W>::value_comp () const {
				return value_compare (key_comp ());
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::iterator NavigableLinkedMap <T, U, W>::find (key_type const & a_key) {
				typename map <key_type, mapped_type, key_compare>::iterator l_keyToValueMapIteratorAtKey = i_keyToValueMap->find (a_key);
				if (l_keyToValueMapIteratorAtKey == i_keyToValueMap->end ()) {
					return iterator (*this, nullptr);
				}
				else {
					return iterator (*this, & (l_keyToValueMapIteratorAtKey->first));
				}
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::const_iterator NavigableLinkedMap <T, U, W>::find (key_type const & a_key) const {
				typename map <key_type, mapped_type, key_compare>::iterator l_keyToValueMapIteratorAtKey = i_keyToValueMap->find (a_key);
				if (l_keyToValueMapIteratorAtKey == i_keyToValueMap->end ()) {
					return const_iterator (*this, nullptr);
				}
				else {
					return const_iterator (*this, & (l_keyToValueMapIteratorAtKey->first));
				}
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::size_type NavigableLinkedMap <T, U, W>::count (key_type const & a_key) const {
				return i_keyToValueMap->count (a_key);
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::iterator NavigableLinkedMap <T, U, W>::lower_bound (key_type const & a_key) {
				key_compare l_keyComparer = this->key_comp ();
				for (value_type const & l_keyValuePair: *this) {
					if (!l_keyComparer (l_keyValuePair.first, a_key)) {
						return iterator (*this, & (l_keyValuePair.first)); 
					}
				}
				return end ();
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::const_iterator NavigableLinkedMap <T, U, W>::lower_bound (key_type const & a_key) const {
				key_compare l_keyComparer = this->key_comp ();
				for (value_type const & l_keyValuePair: *this) {
					if (!l_keyComparer (l_keyValuePair.first, a_key)) {
						return const_iterator (*this, & (l_keyValuePair.first)); 
					}
				}
				return cend ();
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::iterator NavigableLinkedMap <T, U, W>::upper_bound (key_type const & a_key) {
				key_compare l_keyComparer = this->key_comp ();
				for (value_type const & l_keyValuePair: *this) {
					if (l_keyComparer (a_key, l_keyValuePair.first)) {
						return iterator (*this, & (l_keyValuePair.first)); 
					}
				}
				return end ();
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::const_iterator NavigableLinkedMap <T, U, W>::upper_bound (key_type const & a_key) const {
				key_compare l_keyComparer = this->key_comp ();
				for (value_type const & l_keyValuePair: *this) {
					if (l_keyComparer (a_key, l_keyValuePair.first)) {
						return const_iterator (*this, & (l_keyValuePair.first)); 
					}
				}
				return cend ();
			}
			
			template <typename T, typename U, typename W> pair <typename NavigableLinkedMap <T, U, W>::const_iterator, typename NavigableLinkedMap <T, U, W>::const_iterator> NavigableLinkedMap <T, U, W>::equal_range (key_type const & a_key) const {
				const_iterator l_lowerBoundIterator = lower_bound (a_key);
				key_compare l_keyComparer = this->key_comp ();
				if (l_lowerBoundIterator != cend () && ! (l_keyComparer (a_key, l_lowerBoundIterator->first))) {
					const_iterator l_nextToLowerBoundIterator = l_lowerBoundIterator;
					l_nextToLowerBoundIterator ++;
					return pair <const_iterator, const_iterator> (l_lowerBoundIterator, l_nextToLowerBoundIterator);
				}
				else {
					return pair <const_iterator, const_iterator> (l_lowerBoundIterator, l_lowerBoundIterator);
				}
			}
			
			template <typename T, typename U, typename W> pair <typename NavigableLinkedMap <T, U, W>::iterator, typename NavigableLinkedMap <T, U, W>::iterator> NavigableLinkedMap <T, U, W>::equal_range (key_type const & a_key) {
				iterator l_lowerBoundIterator = lower_bound (a_key);
				key_compare l_keyComparer = this->key_comp ();
				if (l_lowerBoundIterator != end () && ! (l_keyComparer (a_key, l_lowerBoundIterator->first))) {
					iterator l_nextToLowerBoundIterator = l_lowerBoundIterator;
					l_nextToLowerBoundIterator ++;
					return pair <iterator, iterator> (l_lowerBoundIterator, l_nextToLowerBoundIterator);
				}
				else {
					return pair <iterator, iterator> (l_lowerBoundIterator, l_lowerBoundIterator);
				}
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::allocator_type NavigableLinkedMap <T, U, W>::get_allocator () const {
				return i_keyToValueMap->get_allocator ();
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::key_type * NavigableLinkedMap <T, U, W>::getFirstKey () const {
				return const_cast <key_type *> (i_firstKey);
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::key_type * NavigableLinkedMap <T, U, W>::getLastKey () const {
				return const_cast <key_type *> (i_lastKey);
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::key_type * NavigableLinkedMap <T, U, W>::getPreviousKey (key_type const & a_key) const {
				try {
					key_type const & l_key = i_keyToValueMap->find (a_key)->first; // the stored key, not any copy of it
					Links & l_links = i_keyToLinksMap->at (&l_key);
					return const_cast <key_type *> (l_links.getPreviousKey ());
				}
				catch (out_of_range & l_exception) {
					return nullptr;
				}
			}
			
			template <typename T, typename U, typename W> typename NavigableLinkedMap <T, U, W>::key_type * NavigableLinkedMap <T, U, W>::getNextKey (key_type const & a_key) const {
				try {
					key_type const & l_key = i_keyToValueMap->find (a_key)->first; // the stored key, not any copy of it
					Links & l_links = i_keyToLinksMap->at (&l_key);
					return const_cast <key_type *> (l_links.getNextKey ());
				}
				catch (out_of_range & l_exception) {
					return nullptr;
				}
			}
		}
	}
}

