#include "theBiasPlanet/unoUtilities/fileConversionsHandling/FilesConverter.hpp"
#include <optional>
#include <com/sun/star/beans/XPropertySet.hpp>
#include <com/sun/star/container/XIndexAccess.hpp>
#include <com/sun/star/container/XNamed.hpp>
#include <com/sun/star/frame/XStorable2.hpp>
#include <com/sun/star/sheet/XSpreadsheetDocument.hpp>
#include <com/sun/star/sheet/XSpreadsheets.hpp>
#include <com/sun/star/uno/Exception.hpp>
#include <com/sun/star/util/XCloseable.hpp>
#include "theBiasPlanet/coreUtilities/collectionsHandling/ListsFactory.hpp"
#include "theBiasPlanet/coreUtilities/constantsGroups/GeneralConstantsConstantsGroup.hpp"
#include "theBiasPlanet/coreUtilities/messagingHandling/Publisher.hpp"
#include "theBiasPlanet/coreUtilities/stringsHandling/StringHandler.hpp"
#include "theBiasPlanet/unoUtilities/constantsGroups/UnoDocumentStoringFilterNamesConstantsGroup.hpp"
#include "theBiasPlanet/unoUtilities/constantsGroups/propertyNamesSets/UnoDocumentOpeningEnumerablePropertyNamesSet.hpp"
#include "theBiasPlanet/unoUtilities/constantsGroups/propertyNamesSets/UnoDocumentStoringEnumerablePropertyNamesSet.hpp"
#include "theBiasPlanet/unoUtilities/constantsGroups/propertyNamesSets/UnoSpreadSheetPropertyNamesSet.hpp"
#include "theBiasPlanet/unoUtilities/propertiesHandling/UnoPropertiesHandler.hpp"
#include "theBiasPlanet/unoUtilities/stringsHandling/UnoExtendedStringHandler.hpp"
#include "theBiasPlanet/unoUtilities/unoDataHandling/UnoDatumConverter.hpp"

using namespace ::com::sun::star::container;
using namespace ::com::sun::star::sheet;
using namespace ::theBiasPlanet::coreUtilities::collectionsHandling;
using namespace ::theBiasPlanet::coreUtilities::constantsGroups;
using namespace ::theBiasPlanet::coreUtilities::messagingHandling;
using namespace ::theBiasPlanet::coreUtilities::stringsHandling;
using namespace ::theBiasPlanet::unoUtilities::constantsGroups;
using namespace ::theBiasPlanet::unoUtilities::constantsGroups::propertyNamesSets;
using namespace ::theBiasPlanet::unoUtilities::propertiesHandling;
using namespace ::theBiasPlanet::unoUtilities::stringsHandling;
using namespace ::theBiasPlanet::unoUtilities::unoDataHandling;

namespace theBiasPlanet {
	namespace unoUtilities {
		namespace fileConversionsHandling {
			FilesConverter::FilesConverter (UnoObjectPointer <UnoDesktop> & a_unoDesktop): i_remoteUnoObjectsContext (a_unoDesktop->getRemoteUnoObjectsContext ()), i_fileOpeningUnoDispatcher (a_unoDesktop->getFileOpeningUnoDispatcher ()) {
			}
			
			FilesConverter::~FilesConverter () {
			}
			
			UnoObjectPointer <RemoteUnoObjectsContext> & FilesConverter::getRemoteUnoObjectsContext () {
				return i_remoteUnoObjectsContext;
			}
			
			bool FilesConverter::convertFile (string const & a_originalFileUrl, optional <string> const & a_originalFilePassword, string const & a_targetFileUrl, Sequence <PropertyValue> const & a_unoDocumentStoringPropertiesSequence, optional <list <UnoDocumentTailor *>> const & a_unoDocumentTailors) {
				bool l_hasSucceeded = false;
				::com::sun::star::util::URL l_originalFileUrlInURL = UnoDatumConverter::getUrlInURL (a_originalFileUrl);
				Sequence <PropertyValue> l_unoDocumentOpeningPropertiesSequence (UnoPropertiesHandler::buildPropertiesSequence (UnoDocumentOpeningEnumerablePropertyNamesSet::c_instance.getValues (), ListsFactory::createList <Any, Any, Any, Any, Any> (Any (true), Any (true), Any (true), Any (true), Any (UnoExtendedStringHandler::getOustring (a_originalFilePassword.has_value () ? *a_originalFilePassword : GeneralConstantsConstantsGroup::c_emptyString)))));
				Any l_underlyingOriginalUnoDocumentInAny = i_fileOpeningUnoDispatcher->dispatchWithReturnValue (l_originalFileUrlInURL, l_unoDocumentOpeningPropertiesSequence);
				if (l_underlyingOriginalUnoDocumentInAny.hasValue ()) {
					UnoObjectPointer <XModel> l_originalUnoDocument (* ( (Reference <XInterface> *) (l_underlyingOriginalUnoDocumentInAny.getValue ())));
					try {
						if (a_unoDocumentTailors.has_value ()) {
							for (UnoDocumentTailor * const l_unoDocumentTailor: *a_unoDocumentTailors) {
								l_unoDocumentTailor->tailor (l_originalUnoDocument);
							}
						}
						l_originalUnoDocument.getPointerInReference <XStorable2> (true)->storeToURL (UnoExtendedStringHandler::getOustring (a_targetFileUrl), a_unoDocumentStoringPropertiesSequence);
						l_hasSucceeded = true;
					}
					catch (Exception & l_exception) {
						Publisher::logErrorInformation (UnoExtendedStringHandler::getString (l_exception.Message));
					}
					catch (...) {
						throw new runtime_error (string ("An exception has occurred."));
					}
					l_originalUnoDocument.getPointerInReference <XCloseable> (true)->close (false); 
				}
				else {
					Publisher::logErrorInformation (StringHandler::format ("### The original file: '%s' cannot be opened.", a_originalFileUrl));
				}
				return l_hasSucceeded;
			}
			
			Sequence <PropertyValue> FilesConverter::createToCsvFileDocumentStoringPropertiesSequence (int const & a_itemsDelimiterCharacterCode, int const & a_textItemQuotationCharacterCode, int const & a_charactersEncodingCode, bool const & a_whetherAllTextItemsAreQuoted, bool const & a_whetherContentsAreExportedAsShown, bool const & a_whetherFormulaThemselvesAreExported) {
				list <string> l_unoDocumentStoringPropertyNames = ListsFactory::createList <string> (
					UnoDocumentStoringEnumerablePropertyNamesSet::c_filterName_string,
					UnoDocumentStoringEnumerablePropertyNamesSet::c_filterData_string,
					UnoDocumentStoringEnumerablePropertyNamesSet::c_overwrites_boolean
				);
				list <Any> l_unoDocumentStoringPropertyValues = ListsFactory::createList <Any> (
					Any (UnoExtendedStringHandler::getOustring (UnoDocumentStoringFilterNamesConstantsGroup::c_toCsvFileFilterName)),
					Any (UnoExtendedStringHandler::getOustring (StringHandler::format (c_csvFormatStringFormat, a_itemsDelimiterCharacterCode, a_textItemQuotationCharacterCode, a_charactersEncodingCode, a_whetherAllTextItemsAreQuoted, a_whetherContentsAreExportedAsShown, a_whetherFormulaThemselvesAreExported))),
					Any (true)
				);
				return UnoPropertiesHandler::buildPropertiesSequence (l_unoDocumentStoringPropertyNames, l_unoDocumentStoringPropertyValues);
			}
			
			bool FilesConverter::convertSpreadSheetsDocumentFileToCsvFiles (string const & a_originalFileUrl, optional <string> const & a_originalFilePassword, string const & a_targetFileUrlBase, Sequence <PropertyValue> const & a_unoDocumentStoringPropertiesSequence, optional <list <UnoDocumentTailor *>> const & a_unoDocumentTailors, int const & a_targetFileNamingRule, bool const & a_hiddenSpreadSheetsAreWritten) {
				bool l_hasSucceeded = false;
				::com::sun::star::util::URL l_originalFileUrlInURL = UnoDatumConverter::getUrlInURL (a_originalFileUrl);
				Sequence <PropertyValue> l_unoDocumentOpeningPropertiesSequence (UnoPropertiesHandler::buildPropertiesSequence (UnoDocumentOpeningEnumerablePropertyNamesSet::c_instance.getValues (), ListsFactory::createList <Any, Any, Any, Any, Any> (Any (true), Any (true), Any (true), Any (true), Any (UnoExtendedStringHandler::getOustring (a_originalFilePassword.has_value () ? *a_originalFilePassword : GeneralConstantsConstantsGroup::c_emptyString)))));
				Any l_underlyingOriginalUnoDocumentInAny = i_fileOpeningUnoDispatcher->dispatchWithReturnValue (l_originalFileUrlInURL, l_unoDocumentOpeningPropertiesSequence);
				if (l_underlyingOriginalUnoDocumentInAny.hasValue ()) {
					UnoObjectPointer <XModel> l_originalUnoDocument (* ( (Reference <XInterface> *) (l_underlyingOriginalUnoDocumentInAny.getValue ())));
					try {
						if (a_unoDocumentTailors.has_value ()) {
							for (UnoDocumentTailor * const l_unoDocumentTailor: *a_unoDocumentTailors) {
								l_unoDocumentTailor->tailor (l_originalUnoDocument);
							}
						}
						if (!(l_originalUnoDocument.getPointerInReference <XSpreadsheetDocument> (true).is ())) {
							Publisher::logErrorInformation ("The document is not any spread sheet.");
						}
						else {
							UnoObjectPointer <XSpreadsheets> l_spreadSheets (l_originalUnoDocument.getPointerInReference <XSpreadsheetDocument> (true)->getSheets ());
							int l_numberOfSheets = l_spreadSheets.getPointerInReference <XIndexAccess> (true)->getCount ();
							UnoObjectPointer <XNamed> l_spreadSheet;
							string l_targetFileNameAugmentation;
							int l_lastPeriodPositionInTargetFileUrlBase = a_targetFileUrlBase.rfind (".");
							string l_augmentedTargetFileUrl;
							bool l_whetherSheetIsVisible = false;
							for (int l_sheetIndex = GeneralConstantsConstantsGroup::c_iterationStartNumber; l_sheetIndex < l_numberOfSheets; l_sheetIndex ++) {
								l_spreadSheet.setPrimaryAddress <XInterface> (* ( (Reference <XInterface> *) l_spreadSheets.getPointerInReference <XIndexAccess> (true)->getByIndex (l_sheetIndex).getValue ()));
								if (a_hiddenSpreadSheetsAreWritten) {
								}
								else {
									l_whetherSheetIsVisible = UnoDatumConverter::convertToBool (l_spreadSheet.getPointerInReference <XPropertySet> ()->getPropertyValue (UnoExtendedStringHandler::getOustring (UnoSpreadSheetPropertyNamesSet::c_whetherIsVisible_boolean))); 
									if (!l_whetherSheetIsVisible) {
										continue;
									}
								}
								if (a_targetFileNamingRule == 0) {
									l_targetFileNameAugmentation = StringHandler::format ("%d", l_sheetIndex);
								}
								else {
									l_targetFileNameAugmentation = UnoExtendedStringHandler::getString (l_spreadSheet->getName ());
								}
								if (l_lastPeriodPositionInTargetFileUrlBase >= GeneralConstantsConstantsGroup::c_iterationStartNumber) {
									l_augmentedTargetFileUrl = StringHandler::format ("%s_%s%s", a_targetFileUrlBase.substr (GeneralConstantsConstantsGroup::c_iterationStartNumber, l_lastPeriodPositionInTargetFileUrlBase), l_targetFileNameAugmentation, a_targetFileUrlBase.substr (l_lastPeriodPositionInTargetFileUrlBase));
								}
								else {
									l_augmentedTargetFileUrl = StringHandler::format ("%s_%s", a_targetFileUrlBase, l_targetFileNameAugmentation);
								}
								if (l_sheetIndex != GeneralConstantsConstantsGroup::c_iterationStartNumber) {
									l_spreadSheets->moveByName (l_spreadSheet->getName (), (short) GeneralConstantsConstantsGroup::c_iterationStartNumber);
								}
								l_originalUnoDocument.getPointerInReference <XStorable2> (true)->storeToURL (UnoExtendedStringHandler::getOustring (l_augmentedTargetFileUrl), a_unoDocumentStoringPropertiesSequence);
							}
							l_hasSucceeded = true;
						}
					}
					catch (Exception & l_exception) {
						Publisher::logErrorInformation (UnoExtendedStringHandler::getString (l_exception.Message));
						throw l_exception;
					}
					catch (...) {
						throw new runtime_error (string ("An exception has occurred."));
					}
					l_originalUnoDocument.getPointerInReference <XCloseable> (true)->close (false); 
				}
				else {
					Publisher::logErrorInformation (StringHandler::format ("### The original file: '%s' cannot be opened.", a_originalFileUrl));
				}
				return l_hasSucceeded;
			}
		}
	}
}

