#include "theBiasPlanet/unoUtilities/pointers/UnoObjectPointer.hpp"
#include <iostream>
#include <com/sun/star/uno/XInterface.hpp>
#include <com/sun/star/lang/WrappedTargetException.hpp>
#include "theBiasPlanet/coreUtilities/messagingHandling/Publisher.hpp"
#include "theBiasPlanet/coreUtilities/stringsHandling/StringHandler.hpp"
#include "theBiasPlanet/unoUtilities/inspectionsHandling/UnoComponentInspector.hpp"
#include "theBiasPlanet/unoUtilities/inspectionsHandling/UnoInterfaceNotImplementedException.hpp"
#include "theBiasPlanet/unoUtilities/stringsHandling/UnoExtendedStringHandler.hpp"
#include "theBiasPlanet/unoUtilities/unoDataHandling/UnoDatumConverter.hpp"

using namespace ::com::sun::star::lang;
using namespace ::theBiasPlanet::coreUtilities::messagingHandling;
using namespace ::theBiasPlanet::coreUtilities::stringsHandling;
using namespace ::theBiasPlanet::unoUtilities::inspectionsHandling;
using namespace ::theBiasPlanet::unoUtilities::stringsHandling;
using namespace ::theBiasPlanet::unoUtilities::unoDataHandling;

namespace theBiasPlanet {
	namespace unoUtilities {
		namespace pointers {
			template <typename T> void UnoObjectPointer <T>::initialize () {
				if (i_primaryAddress != nullptr) {
					Publisher::logDebugInformation ("### An instance of 'UnoObjectPointer' is being initialized.");
					i_primaryAddress->acquire ();
					if (UnoObjectPointersInformationDumpCondition::c_unoObjectsInformationDumpFilePathString.has_value ()) {
						try {
							Publisher::appendToFile (* (UnoObjectPointersInformationDumpCondition::c_unoObjectsInformationDumpFilePathString), StringHandler::format ("### %s -> %s", string (typeid (T).name ()), StringHandler::getString (getUnoTypeNames (), string (", "))));
							ostringstream l_propertyNameValuePairsStringBuilder;
							NavigableLinkedMap <string, Any> l_propertyNameToValueMap (getPropertyNameToValueMap ());
							bool l_is1stItem = true;
							for (NavigableLinkedMap <string, Any>::value_type l_propertyNameToValueMapEntry: l_propertyNameToValueMap) {
								if (l_is1stItem) {
									l_is1stItem = false;
								}
								else {
									l_propertyNameValuePairsStringBuilder << string (", ");
								}
								l_propertyNameValuePairsStringBuilder << string (l_propertyNameToValueMapEntry.first);
								l_propertyNameValuePairsStringBuilder << string (": ");
								l_propertyNameValuePairsStringBuilder << UnoDatumConverter::convertToNonUnoDatum (l_propertyNameToValueMapEntry.second);
							}
							Publisher::appendToFile (* (UnoObjectPointersInformationDumpCondition::c_unoObjectsInformationDumpFilePathString), StringHandler::format ("### %s -> %s", string (typeid (T).name ()), l_propertyNameValuePairsStringBuilder.str ()));
						}
						catch (UnoInterfaceNotImplementedException const & l_exception) {
							Publisher::appendToFile (* (UnoObjectPointersInformationDumpCondition::c_unoObjectsInformationDumpFilePathString), StringHandler::format ("### %s -> %s", string (typeid (T).name ()), StringHandler::format ("%s: %s", string ("UnoInterfaceNotImplementedException"), string (l_exception.what ()))));
						}
						catch (WrappedTargetException const & l_exception) {
							Publisher::appendToFile (* (UnoObjectPointersInformationDumpCondition::c_unoObjectsInformationDumpFilePathString), StringHandler::format ("### %s -> %s", string (typeid (T).name ()), StringHandler::format ("%s: %s", string ("WrappedTargetException"), UnoExtendedStringHandler::getString (l_exception.Message))));
						}
					}
				}
			}
			
			template <typename T> void UnoObjectPointer <T>::setUnoObjectsInformationDumpFilePathString (string const & a_unoObjectsInformationDumpFilePathString) {
				UnoObjectPointersInformationDumpCondition::c_unoObjectsInformationDumpFilePathString = a_unoObjectsInformationDumpFilePathString;
			}
			
			template <typename T> void UnoObjectPointer <T>::unoInterfaceNameToPointerInReferenceMapDeleter (map <string const, void * const> * a_pointer) {
				map <string const, void * const>::iterator l_unoInterfaceNameToPointerInReferenceMapIterator (a_pointer->begin ());
				map <string const, void * const>::iterator l_unoInterfaceNameToPointerInReferenceMapIteratorAtEnd (a_pointer->end ());
				for (; l_unoInterfaceNameToPointerInReferenceMapIterator != l_unoInterfaceNameToPointerInReferenceMapIteratorAtEnd;) {
					string l_unoInterfaceName (l_unoInterfaceNameToPointerInReferenceMapIterator->first);
					// * In fact, it isn't 'Reference <XInterface> *', but practically, that doesn't seem to matter; if it turns out to matter, the address will have to be cast to the correct type.
					Publisher::logDebugInformation ("### An instance of 'Reference' in an instance of 'UnoObjectPointer' is being disposed.");
					delete (Reference <XInterface> *) (l_unoInterfaceNameToPointerInReferenceMapIterator->second);
					l_unoInterfaceNameToPointerInReferenceMapIterator = a_pointer->erase (l_unoInterfaceNameToPointerInReferenceMapIterator);
				}
			}
			
			template <typename T> UnoObjectPointer <T>::UnoObjectPointer () {
			}
			
			template <typename T> UnoObjectPointer <T>::UnoObjectPointer (T * const a_primaryAddress): i_primaryAddress (a_primaryAddress), i_unoInterfaceNameToPointerInReferenceMap (new map <string const, void * const> (), &unoInterfaceNameToPointerInReferenceMapDeleter) {
				initialize ();
			}
			
			template <typename T> UnoObjectPointer <T>::UnoObjectPointer (Reference <T> const & a_primaryAddressPointer): UnoObjectPointer (a_primaryAddressPointer.get ()) {
			}
			
			template <typename T> template <typename U> UnoObjectPointer <T>::UnoObjectPointer (Reference <U> const & a_sourcePointer, typename enable_if <!is_same <T, U>::value>::type * a_dummy): UnoObjectPointer <T> (Reference <T> (a_sourcePointer, UNO_QUERY)) {
			}
			
			template <typename T> template <typename U> UnoObjectPointer <T>::UnoObjectPointer (UnoObjectPointer <U> & a_sourcePointer, bool const & a_targetAddressIsQueried, typename enable_if <!is_same <T, U>::value>::type * a_dummy): i_primaryAddress (a_sourcePointer.template getPointerInReference <T> (a_targetAddressIsQueried).get ()), i_unoInterfaceNameToPointerInReferenceMap (a_sourcePointer.getUnoInterfaceNameToPointerInReferenceMap ()) {
				initialize ();
			}
			
			//template <typename T> template <typename U> UnoObjectPointer <T>::UnoObjectPointer (UnoObjectPointer <U> && a_sourcePointer, bool const & a_targetAddressIsQueried, typename enable_if <!is_same <T, U>::value>::type * a_dummy): i_primaryAddress (a_sourcePointer.template getPointerInReference <T> (a_targetAddressIsQueried)), i_unoInterfaceNameToPointerInReferenceMap (a_sourcePointer.i_unoInterfaceNameToPointerInReferenceMap) {
			template <typename T> template <typename U> UnoObjectPointer <T>::UnoObjectPointer (UnoObjectPointer <U> && a_sourcePointer, bool const & a_targetAddressIsQueried, typename enable_if <!is_same <T, U>::value>::type * a_dummy): UnoObjectPointer (a_sourcePointer, a_targetAddressIsQueried) {
				initialize ();
			}
			
			template <typename T> UnoObjectPointer <T>::UnoObjectPointer (UnoObjectPointer <T> const & a_sourcePointer): i_primaryAddress (a_sourcePointer.i_primaryAddress), i_unoInterfaceNameToPointerInReferenceMap (a_sourcePointer.i_unoInterfaceNameToPointerInReferenceMap) {
				Publisher::logDebugInformation ("### An instance of 'UnoObjectPointer' is being copied.");
				initialize ();
			}
			
			template <typename T> UnoObjectPointer <T> & UnoObjectPointer <T>::operator = (UnoObjectPointer <T> const & a_sourcePointer) {
				Publisher::logDebugInformation ("### An instance of 'UnoObjectPointer' is being assigned.");
				clear ();
				i_primaryAddress = a_sourcePointer.i_primaryAddress;
				initialize ();
				i_unoInterfaceNameToPointerInReferenceMap = a_sourcePointer.i_unoInterfaceNameToPointerInReferenceMap;
				return *this;
			}
			
			template <typename T> UnoObjectPointer <T>::~UnoObjectPointer () {
				clear ();
			}
			
			template <typename T> list <string> UnoObjectPointer <T>::getUnoTypeNames () {
				return UnoComponentInspector::getUnoTypeNames (i_primaryAddress);
			}
			
			template <typename T> NavigableLinkedMap <string, Any> UnoObjectPointer <T>::getPropertyNameToValueMap () {
				return UnoComponentInspector::getPropertyNameToValueMap (i_primaryAddress);
			}
			
			template <typename T> bool UnoObjectPointer <T>::isEmpty () {
				return i_primaryAddress == nullptr;
			}
			
			template <typename T> void UnoObjectPointer <T>::setPrimaryAddress (T * const a_primaryAddress) {
				clear ();
				i_primaryAddress = a_primaryAddress;
				initialize ();
				i_unoInterfaceNameToPointerInReferenceMap.reset (new map <string const, void * const> (), &unoInterfaceNameToPointerInReferenceMapDeleter);
			}
			
			template <typename T> void UnoObjectPointer <T>::setPrimaryAddress (Reference <T> const & a_primaryAddressPointer) {
				setPrimaryAddress (a_primaryAddressPointer.get ());
			}
			
			template <typename T> template <typename U>  void UnoObjectPointer <T>::setPrimaryAddress (Reference <U> const & a_sourcePointer, typename enable_if <!is_same <T, U>::value>::type * a_dummy) {
				setPrimaryAddress (Reference <T> (a_sourcePointer, UNO_QUERY));
			}
			
			template <typename T> void UnoObjectPointer <T>::setPrimaryAddress (UnoObjectPointer <T> & a_sourcePointer) {
				clear ();
				i_primaryAddress = a_sourcePointer.getPointerInReference ().get ();
				initialize ();
				i_unoInterfaceNameToPointerInReferenceMap = a_sourcePointer.getUnoInterfaceNameToPointerInReferenceMap ();
			}
			
			template <typename T> template <typename U> void UnoObjectPointer <T>::setPrimaryAddress (UnoObjectPointer <U> & a_sourcePointer, bool const & a_targetAddressIsQueried, typename enable_if <!is_same <T, U>::value>::type * a_dummy) {
				clear ();
				i_primaryAddress = a_sourcePointer.template getPointerInReference <T> (a_targetAddressIsQueried).get ();
				initialize ();
				i_unoInterfaceNameToPointerInReferenceMap = a_sourcePointer.getUnoInterfaceNameToPointerInReferenceMap ();
			}
			
			template <typename T> template <typename U> void UnoObjectPointer <T>::setPrimaryAddress (UnoObjectPointer <U> && a_sourcePointer, bool const & a_targetAddressIsQueried, typename enable_if <!is_same <T, U>::value>::type * a_dummy) {
				setPrimaryAddress (a_sourcePointer, a_targetAddressIsQueried);
			}
			
			template <typename T> void UnoObjectPointer <T>::clear () {
				if (i_primaryAddress != nullptr) {
					Publisher::logDebugInformation ("### An instance of 'UnoObjectPointer' is being cleared.");
					i_primaryAddress->release ();
					/*
					map <string const, void * const>::iterator l_unoInterfaceNameToPointerInReferenceMapIterator (i_unoInterfaceNameToPointerInReferenceMap.begin ());
					map <string const, void * const>::iterator l_unoInterfaceNameToPointerInReferenceMapIteratorAtEnd (i_unoInterfaceNameToPointerInReferenceMap.end ());
					for (; l_unoInterfaceNameToPointerInReferenceMapIterator != l_unoInterfaceNameToPointerInReferenceMapIteratorAtEnd;) {
						string l_unoInterfaceName (l_unoInterfaceNameToPointerInReferenceMapIterator->first);
						// * In fact, it isn't 'Reference <XInterface> *', but practically, that doesn't seem to matter; if it turns out to matter, the address will have to be cast to the correct type.
						Publisher::logDebugInformation ("### An instance of 'Reference' in an instance of 'UnoObjectPointer' is being disposed.");
						delete (Reference <XInterface> *) (l_unoInterfaceNameToPointerInReferenceMapIterator->second);
						l_unoInterfaceNameToPointerInReferenceMapIterator = i_unoInterfaceNameToPointerInReferenceMap.erase (l_unoInterfaceNameToPointerInReferenceMapIterator);
					}
					*/
					i_unoInterfaceNameToPointerInReferenceMap.reset ();
					i_primaryAddress = nullptr;
				}
			}
			
			template <typename T> T & UnoObjectPointer <T>::operator * () {
				return *i_primaryAddress;
			}
			
			template <typename T> T * const UnoObjectPointer <T>::operator -> () {
				return i_primaryAddress;
			}
			
			template <typename T> Reference <T> & UnoObjectPointer <T>::getPointerInReference () {
				void * l_pointerInReference = nullptr;
				string l_unoInterfaceName (typeid (T).name ());
				try {
					l_pointerInReference = i_unoInterfaceNameToPointerInReferenceMap.get ()->at (l_unoInterfaceName);
				}
				catch (out_of_range & l_exception) {
					Publisher::logDebugInformation ("### An instance of 'Reference' in an instance of 'UnoObjectPointer' is being created.");
					l_pointerInReference = new Reference <T> ( (T *) i_primaryAddress);
					i_unoInterfaceNameToPointerInReferenceMap.get ()->insert ( {l_unoInterfaceName, l_pointerInReference});
				}
				return * ((Reference <T> *) l_pointerInReference);
			}
			
			template <typename T> template <typename U> Reference <U> & UnoObjectPointer <T>::getPointerInReference (bool const & a_pointerIsQueried, typename enable_if <!is_same <T, U>::value>::type * a_dummy) {
				void * l_pointerInReference = nullptr;
				string l_unoInterfaceName (typeid (U).name ());
				try {
					l_pointerInReference = i_unoInterfaceNameToPointerInReferenceMap.get ()->at (l_unoInterfaceName);
				}
				catch (out_of_range & l_exception) {
					Publisher::logDebugInformation ("### An instance of 'Reference' in an instance of 'UnoObjectPointer' is being created.");
					if (!a_pointerIsQueried) {
						l_pointerInReference = new Reference <U> ( (U *) i_primaryAddress);
					}
					else {
						l_pointerInReference = new Reference <U> ( (U *) i_primaryAddress, UNO_QUERY);
					}
					i_unoInterfaceNameToPointerInReferenceMap.get ()->insert ( {l_unoInterfaceName, l_pointerInReference});
				}
				return * ((Reference <U> *) l_pointerInReference);
			}
			
			template <typename T> shared_ptr <map <string const, void * const>> const & UnoObjectPointer <T>::getUnoInterfaceNameToPointerInReferenceMap () {
				return i_unoInterfaceNameToPointerInReferenceMap;
			}
		}
	}
}

