#include "theBiasPlanet/unoUtilities/unoObjectsContexts/UnoObjectsContext.hpp"
#include <com/sun/star/container/XNameContainer.hpp>
#include "theBiasPlanet/unoUtilities/collectionsHandling/UnoSequenceHandler.hpp"
#include "theBiasPlanet/unoUtilities/constantsGroups/propertyNamesSets/UnoObjectsContextPropertyNamesSet.hpp"
#include "theBiasPlanet/unoUtilities/stringsHandling/UnoExtendedStringHandler.hpp"

using namespace ::com::sun::star::container;
using namespace ::theBiasPlanet::unoUtilities::collectionsHandling;
using namespace ::theBiasPlanet::unoUtilities::constantsGroups::propertyNamesSets;
using namespace ::theBiasPlanet::unoUtilities::stringsHandling;

namespace theBiasPlanet {
	namespace unoUtilities {
		namespace unoObjectsContexts {
			UnoObjectsContext::UnoObjectsContext (UnoObjectPointer <XComponentContext> const & a_underlyingUnoObject, string const & a_unoConnectionIdentification): ImplInheritanceHelper1 <UnoComponentBase, XComponentContext> (string ("::theBiasPlanet::unoUtilities::unoObjectsContexts::UnoObjectsContext")), i_underlyingUnoObject (a_underlyingUnoObject) {
				addLocalProperty (UnoObjectsContextPropertyNamesSet::c_unoConnectionIdentification_string, new string (a_unoConnectionIdentification));
				i_unoGlobalServicesManager.setPrimaryAddress (i_underlyingUnoObject->getServiceManager ());
			}
			
			UnoObjectsContext::~UnoObjectsContext () {
				string * l_unoConnectionIdentification = (string *) getLocalPropertyValue (UnoObjectsContextPropertyNamesSet::c_unoConnectionIdentification_string);
				if (l_unoConnectionIdentification != nullptr) {
					delete l_unoConnectionIdentification;
					removeLocalProperty (UnoObjectsContextPropertyNamesSet::c_unoConnectionIdentification_string);
				}
			}
			
			UnoObjectPointer <XComponentContext> & UnoObjectsContext::getUnderlyingUnoObject () {
				return i_underlyingUnoObject;
			}
			
			Any UnoObjectsContext::getValueByName (OUString const & a_propertyName) {
				return i_underlyingUnoObject->getValueByName (a_propertyName);
			}
			
			void UnoObjectsContext::addExtraGlobalProperty (string const & a_propertyName, Any const & a_propertyValue) {
				i_underlyingUnoObject.getPointerInReference <XNameContainer> ()->insertByName (UnoExtendedStringHandler::getOustring (a_propertyName), a_propertyValue);
			}
			
			void UnoObjectsContext::removeExtraGlobalProperty (string const & a_propertyName) {
				i_underlyingUnoObject.getPointerInReference <XNameContainer> ()->removeByName (UnoExtendedStringHandler::getOustring (a_propertyName));
			}
			
			void * const UnoObjectsContext::getLocalPropertyValue (string const & a_propertyName) {
				return i_localPropertyNameToValueMap [a_propertyName];
			}
			
			void UnoObjectsContext::addLocalProperty (string const & a_propertyName, void * a_propertyValue) {
				i_localPropertyNameToValueMap [a_propertyName] = a_propertyValue;
			}
			
			void UnoObjectsContext::removeLocalProperty (string const & a_propertyName) {
				i_localPropertyNameToValueMap.erase (i_localPropertyNameToValueMap.find (a_propertyName));
			}
			
			Reference <XMultiComponentFactory> UnoObjectsContext::getServiceManager () {
				return i_unoGlobalServicesManager.getPointerInReference ();
			}
			
			bool UnoObjectsContext::isFromSameOrigin (UnoObjectPointer <UnoObjectsContext> & a_comparedUnoObjectsContext) {
				if (a_comparedUnoObjectsContext.isEmpty ()) {
					return false;
				}
				string const * const l_unoConnectionIdentification = (string *) getLocalPropertyValue (UnoObjectsContextPropertyNamesSet::c_unoConnectionIdentification_string);
				if (l_unoConnectionIdentification == nullptr) {
					return false;
				}
				if (*l_unoConnectionIdentification == * ( (string *) (a_comparedUnoObjectsContext->getLocalPropertyValue (UnoObjectsContextPropertyNamesSet::c_unoConnectionIdentification_string)))) {
					return true;
				}
				return false;
			}
			
			void UnoObjectsContext::getUnoServiceInstance (UnoObjectPointer <XInterface> & a_targetUnoObjectPointer, string const & a_unoServiceName, optional <list <Any>> const & a_instantiationArguments) {
				if (!a_instantiationArguments.has_value ()) {
					a_targetUnoObjectPointer.setPrimaryAddress (getServiceManager ()->createInstanceWithContext (UnoExtendedStringHandler::getOustring (a_unoServiceName), i_underlyingUnoObject.getPointerInReference ()));
				}
				else {
					Sequence <Any> l_unoServiceInstantiationArgumentsSequence;
					a_targetUnoObjectPointer.setPrimaryAddress (getServiceManager ()->createInstanceWithArgumentsAndContext (UnoExtendedStringHandler::getOustring (a_unoServiceName), *(UnoSequenceHandler::fillSequence <Any, Any> (&l_unoServiceInstantiationArgumentsSequence, a_instantiationArguments.value ())), i_underlyingUnoObject.getPointerInReference ()));
				}
			}
			
			UnoObjectPointer <XInterface> UnoObjectsContext::getUnoServiceInstance (string const & a_unoServiceName, optional <list <Any>> const & a_instantiationArguments) {
				UnoObjectPointer <XInterface> l_targetUnoObjectPointer (nullptr);
				getUnoServiceInstance (l_targetUnoObjectPointer, a_unoServiceName, a_instantiationArguments);
				return l_targetUnoObjectPointer;
			}
		}
	}
}

