namespace theBiasPlanet {
	namespace unoUtilities {
		namespace cryptographyHandling {
			using System;
			using System.Text;
			using theBiasPlanet.coreUtilities.constantsGroups;
			
			public class MicrosoftPasswordsHasher {
				private const Int32 c_numberOfInitializationCodes = 15;
				private static readonly UInt16 [] c_initializationCodes = new UInt16 [] {(UInt16) 0xE1F0, (UInt16) 0x1D0F, (UInt16) 0xCC9C, (UInt16) 0x84C0, (UInt16) 0x110C, (UInt16) 0x0E10, (UInt16) 0xF1CE, (UInt16) 0x313E, (UInt16) 0x1872, (UInt16) 0xE139, (UInt16) 0xD40F, (UInt16) 0x84F9, (UInt16) 0x280C, (UInt16) 0xA96A, (UInt16) 0x4EC3};
				private const Int32 c_numberOfEncryptionMatrixColumns = 7;
				private static readonly UInt16 [, ] c_encryptionMatrix = new UInt16 [, ] {
					{ (UInt16) 0xAEFC, (UInt16) 0x4DD9, (UInt16) 0x9BB2, (UInt16) 0x2745, (UInt16) 0x4E8A, (UInt16) 0x9D14, (UInt16) 0x2A09},
					{ (UInt16) 0x7B61, (UInt16) 0xF6C2, (UInt16) 0xFDA5, (UInt16) 0xEB6B, (UInt16) 0xC6F7, (UInt16) 0x9DCF, (UInt16) 0x2BBF},
					{ (UInt16) 0x4563, (UInt16) 0x8AC6, (UInt16) 0x05AD, (UInt16) 0x0B5A, (UInt16) 0x16B4, (UInt16) 0x2D68, (UInt16) 0x5AD0},
					{ (UInt16) 0x0375, (UInt16) 0x06EA, (UInt16) 0x0DD4, (UInt16) 0x1BA8, (UInt16) 0x3750, (UInt16) 0x6EA0, (UInt16) 0xDD40},
					{ (UInt16) 0xD849, (UInt16) 0xA0B3, (UInt16) 0x5147, (UInt16) 0xA28E, (UInt16) 0x553D, (UInt16) 0xAA7A, (UInt16) 0x44D5},
					{ (UInt16) 0x6F45, (UInt16) 0xDE8A, (UInt16) 0xAD35, (UInt16) 0x4A4B, (UInt16) 0x9496, (UInt16) 0x390D, (UInt16) 0x721A},
					{ (UInt16) 0xEB23, (UInt16) 0xC667, (UInt16) 0x9CEF, (UInt16) 0x29FF, (UInt16) 0x53FE, (UInt16) 0xA7FC, (UInt16) 0x5FD9},
					{ (UInt16) 0x47D3, (UInt16) 0x8FA6, (UInt16) 0x8FA6, (UInt16) 0x1EDA, (UInt16) 0x3DB4, (UInt16) 0x7B68, (UInt16) 0xF6D0},
					{ (UInt16) 0xB861, (UInt16) 0x60E3, (UInt16) 0xC1C6, (UInt16) 0x93AD, (UInt16) 0x377B, (UInt16) 0x6EF6, (UInt16) 0xDDEC},
					{ (UInt16) 0x45A0, (UInt16) 0x8B40, (UInt16) 0x06A1, (UInt16) 0x0D42, (UInt16) 0x1A84, (UInt16) 0x3508, (UInt16) 0x6A10},
					{ (UInt16) 0xAA51, (UInt16) 0x4483, (UInt16) 0x8906, (UInt16) 0x022D, (UInt16) 0x045A, (UInt16) 0x08B4, (UInt16) 0x1168},
					{ (UInt16) 0x76B4, (UInt16) 0xED68, (UInt16) 0xCAF1, (UInt16) 0x85C3, (UInt16) 0x1BA7, (UInt16) 0x374E, (UInt16) 0x6E9C},
					{ (UInt16) 0x3730, (UInt16) 0x6E60, (UInt16) 0xDCC0, (UInt16) 0xA9A1, (UInt16) 0x4363, (UInt16) 0x86C6, (UInt16) 0x1DAD},
					{ (UInt16) 0x3331, (UInt16) 0x6662, (UInt16) 0xCCC4, (UInt16) 0x89A9, (UInt16) 0x0373, (UInt16) 0x06E6, (UInt16) 0x0DCC},
					{ (UInt16) 0x1021, (UInt16) 0x2042, (UInt16) 0x4084, (UInt16) 0x8108, (UInt16) 0x1231, (UInt16) 0x2462, (UInt16) 0x48C4}
				};
				
				public static Int32 hashIn32bits (String a_originalDatum) {
					UInt32 l_hash = 0;
					Int32 l_originalDatumLength = a_originalDatum.Length;
					if (l_originalDatumLength > 0) {
						if (l_originalDatumLength > c_numberOfInitializationCodes) {
					   		l_originalDatumLength = c_numberOfInitializationCodes;
						}
						UInt16 l_highHash = c_initializationCodes [l_originalDatumLength - 1];
						UInt16 l_lowHash = 0;
						Char l_character = (Char) 0x0000;
						Byte l_byte = 0x00;
						Byte l_highByte = 0x00;
						Byte l_lowByte = 0x00;
						for (Int32 l_characterIndex = GeneralConstantsConstantsGroup.c_iterationStartNumber; l_characterIndex < l_originalDatumLength; l_characterIndex ++) {
							l_character = a_originalDatum [l_characterIndex];
							l_highByte = (Byte) (l_character >> 8);
							l_lowByte = (Byte) (l_character & 0xFF);
							l_byte = l_lowByte != 0x00 ? l_lowByte : l_highByte;
							for (Int32 l_matrixcolumnIndex = GeneralConstantsConstantsGroup.c_iterationStartNumber; l_matrixcolumnIndex < c_numberOfEncryptionMatrixColumns; l_matrixcolumnIndex ++) {
								if ((l_byte & (1 << l_matrixcolumnIndex)) != 0) {
									l_highHash = (UInt16) (l_highHash ^ c_encryptionMatrix [c_numberOfInitializationCodes - l_originalDatumLength + l_characterIndex, l_matrixcolumnIndex]);
								}
							}
							l_character = a_originalDatum [l_originalDatumLength -1 - l_characterIndex];
							l_highByte = (Byte) (l_character >> 8);
							l_lowByte = (Byte) (l_character & 0xFF);
							l_byte = l_lowByte != 0x00 ? l_lowByte : l_highByte;
							l_lowHash = (UInt16) (( ( (l_lowHash >> 14) & 0x0001 ) | ( (l_lowHash << 1) & 0x7FFF)) ^ l_byte);
						}
						l_lowHash =  (UInt16) (( ( (l_lowHash >> 14) & 0x0001) | ( (l_lowHash << 1) & 0x7FFF)) ^ l_originalDatumLength ^ 0xCE4B);
						l_hash = (UInt32) ((l_highHash << 16) | l_lowHash);
					}
					return (Int32) l_hash;
				}
				
				public static Int16 hashIn16bits (String a_originalDatum) {
					UInt16 l_hash = 0;
					Byte [] l_originalDatumUtf8BytesArray = null;
					try {
						l_originalDatumUtf8BytesArray = Encoding.UTF8.GetBytes (a_originalDatum);
					}
					catch (Exception) {
						// Is supposed to never happen.
					}
					Int32 l_originalDatumUtf8BytesArrayLength = l_originalDatumUtf8BytesArray.Length;
					if (l_originalDatumUtf8BytesArrayLength <= Int16.MaxValue) {
						for (Int32 l_byteIndex = l_originalDatumUtf8BytesArrayLength - 1; l_byteIndex >= GeneralConstantsConstantsGroup.c_iterationStartNumber; l_byteIndex --) {
							l_hash = (UInt16) (( (l_hash >> 14) & 0x01) | ( (l_hash << 1) & 0x7FFF));
							l_hash ^= l_originalDatumUtf8BytesArray [l_byteIndex];
						}
						l_hash = (UInt16) (( (l_hash >> 14) & 0x01) | ( (l_hash << 1) & 0x7FFF));
						l_hash ^= (0x8000 | ('N' << 8) | 'K');
						l_hash ^= (UInt16) l_originalDatumUtf8BytesArrayLength;
					}
					return (Int16)l_hash;
				}
			}
		}
	}
}

