namespace theBiasPlanet {
	namespace unoUtilities {
		namespace fileConversionsHandling {
			using System;
			using System.Collections.Generic;
			using uno;
			using unoidl.com.sun.star.beans;
			using unoidl.com.sun.star.container;
			using unoidl.com.sun.star.frame;
			using unoidl.com.sun.star.lang;
			using unoidl.com.sun.star.sheet;
			using unoidl.com.sun.star.util;
			using theBiasPlanet.coreUtilities.collectionsHandling;
			using theBiasPlanet.coreUtilities.constantsGroups;
			using theBiasPlanet.coreUtilities.messagingHandling;
			using theBiasPlanet.unoUtilities.constantsGroups;
			using theBiasPlanet.unoUtilities.constantsGroups.propertyNamesSets;
			using theBiasPlanet.unoUtilities.displayElements;
			using theBiasPlanet.unoUtilities.documentsHandling;
			using theBiasPlanet.unoUtilities.pointers;
			using theBiasPlanet.unoUtilities.propertiesHandling;
			using theBiasPlanet.unoUtilities.unoDataHandling;
			using theBiasPlanet.unoUtilities.unoObjectsContexts;
			
			public class FilesConverter {
				private static readonly String c_csvFormatStringFormat = "{0:d},{1:d},{2:d},1,,0,{3:s},false,{4:s},{5:s},false"; // the CSV items delimiter character code (can be specified like 'Character.codePointAt ("\t", 0)'), the CSV text items quotation character code (can be specified like 'Character.codePointAt ("\"", 0)'), the encoding code (76 -> UTF-8, 65535 -> UCS-2, 65534 -> UCS-4, 11 -> US-ASCII, 69 -> EUC_JP, 64 -> SHIFT_JIS), whether all the text items are quoted, whether the contents are exported as shown, whether the formula themselves are exported
				private RemoteUnoObjectsContext i_remoteUnoObjectsContext = null;
				private UnoObjectPointer <XSynchronousDispatch> i_fileOpeningUnoDispatcher = null;
				
				public FilesConverter (UnoDesktop a_unoDesktop) {
					i_remoteUnoObjectsContext = a_unoDesktop.getRemoteUnoObjectsContext ();
					i_fileOpeningUnoDispatcher = a_unoDesktop.getFileOpeningDispatcher ();
				}
				
				public RemoteUnoObjectsContext getRemoteUnoObjectsContext () {
					return i_remoteUnoObjectsContext;
				}
				
				public Boolean convertFile (String a_originalFileUrl, String a_originalFilePassword, String a_targetFileUrl, PropertyValue [] a_unoDocumentStoringPropertiesArray, List <UnoDocumentTailor> a_unoDocumentTailors) {
					unoidl.com.sun.star.util.URL l_originalFileUrlInURL = UnoDatumConverter.getUrlInURL (a_originalFileUrl);
					PropertyValue [] l_unoDocumentOpeningPropertiesArray = UnoPropertiesHandler.buildPropertiesArray (UnoDocumentOpeningEnumerablePropertyNamesSet.c_instance.getValues (), ListsFactory.createList <Any> (UnoDatumConverter.getAny (true), UnoDatumConverter.getAny (true), UnoDatumConverter.getAny (true), UnoDatumConverter.getAny (true), UnoDatumConverter.getAny (a_originalFilePassword != null ? a_originalFilePassword: "")));
					Any l_underlyingOriginalUnoDocumentInAny = (Any) i_fileOpeningUnoDispatcher.getAddress ().dispatchWithReturnValue (l_originalFileUrlInURL, l_unoDocumentOpeningPropertiesArray);
					Boolean l_hasSucceeded = false;
					if (l_underlyingOriginalUnoDocumentInAny.hasValue ()) {
						UnoObjectPointer <XModel> l_originalUnoDocument = new UnoObjectPointer <XModel> ( (Object) l_underlyingOriginalUnoDocumentInAny.Value);
						try {
							if (a_unoDocumentTailors != null) {
								foreach (UnoDocumentTailor l_unoDocumentTailor in a_unoDocumentTailors) {
									l_unoDocumentTailor.tailor (l_originalUnoDocument);
								}
							}
							l_originalUnoDocument.getAddress <XStorable2> ().storeToURL (a_targetFileUrl, a_unoDocumentStoringPropertiesArray);
							l_hasSucceeded = true;
						}
						catch (Exception l_exception) {
							throw l_exception;
						}
						l_originalUnoDocument.getAddress <XCloseable> ().close (false); 
					}
					else {
						Publisher.logErrorInformation (String.Format ("### The original file: '{0:s}' cannot be opened.", a_originalFileUrl));
					}
					return l_hasSucceeded;
				}
				
				public static PropertyValue [] createToCsvFileDocumentStoringPropertiesArray (Int32 a_itemsDelimiterCharacterCode, Int32 a_textItemQuotationCharacterCode, Int32 a_charactersEncodingCode, Boolean a_whetherAllTextItemsAreQuoted, Boolean a_whetherContentsAreExportedAsShown, Boolean a_whetherFormulaThemselvesAreExported) {
					List <String> l_unoDocumentStoringPropertyNames = ListsFactory.createList <String> (
						UnoDocumentStoringEnumerablePropertyNamesSet.c_filterName_string,
						UnoDocumentStoringEnumerablePropertyNamesSet.c_filterData_string,
						UnoDocumentStoringEnumerablePropertyNamesSet.c_overwrites_boolean
					);
					List <Any> l_unoDocumentStoringPropertyValues = ListsFactory.createList <Any> (
						UnoDatumConverter.getAny (UnoDocumentStoringFilterNamesConstantsGroup.c_toCsvFileFilterName),
						UnoDatumConverter.getAny (String.Format (c_csvFormatStringFormat, a_itemsDelimiterCharacterCode, a_textItemQuotationCharacterCode, a_charactersEncodingCode, a_whetherAllTextItemsAreQuoted.ToString ().ToLower (), a_whetherContentsAreExportedAsShown.ToString ().ToLower (), a_whetherFormulaThemselvesAreExported.ToString ().ToLower ())),
						UnoDatumConverter.getAny (true)
					);
					return UnoPropertiesHandler.buildPropertiesArray (l_unoDocumentStoringPropertyNames, l_unoDocumentStoringPropertyValues);
				}
				
				// a_targetFileNamingRule: 0 -> augmented by the sheet index, 1 -> augmented by the sheet name
				public Boolean convertSpreadSheetsDocumentFileToCsvFiles (String a_originalFileUrl, String a_originalFilePassword, String a_targetFileUrlBase, PropertyValue [] a_unoDocumentStoringPropertiesArray, List <UnoDocumentTailor> a_unoDocumentTailors, Int32 a_targetFileNamingRule, Boolean a_hiddenSpreadSheetsAreWritten) {
					unoidl.com.sun.star.util.URL l_originalFileUrlInURL = UnoDatumConverter.getUrlInURL (a_originalFileUrl);
					PropertyValue [] l_unoDocumentOpeningPropertiesArray = UnoPropertiesHandler.buildPropertiesArray (UnoDocumentOpeningEnumerablePropertyNamesSet.c_instance.getValues (), ListsFactory.createList <Any> (UnoDatumConverter.getAny (true), UnoDatumConverter.getAny (true), UnoDatumConverter.getAny (true), UnoDatumConverter.getAny (true), UnoDatumConverter.getAny (a_originalFilePassword != null ? a_originalFilePassword: "")));
					Any l_underlyingOriginalUnoDocumentInAny = (Any) i_fileOpeningUnoDispatcher.getAddress ().dispatchWithReturnValue (l_originalFileUrlInURL, l_unoDocumentOpeningPropertiesArray);
					Boolean l_hasSucceeded = false;
					if (l_underlyingOriginalUnoDocumentInAny.hasValue ()) {
						UnoObjectPointer <XModel> l_originalUnoDocument = new UnoObjectPointer <XModel> ( (Object) l_underlyingOriginalUnoDocumentInAny.Value);
						try {
							if (a_unoDocumentTailors != null) {
								foreach (UnoDocumentTailor l_unoDocumentTailor in a_unoDocumentTailors) {
									l_unoDocumentTailor.tailor (l_originalUnoDocument);
								}
							}
							if (l_originalUnoDocument.getAddress <XSpreadsheetDocument> () == null) {
								Publisher.logErrorInformation ("The document is not any spread sheet.");
								return false;
							}
							else {
								UnoObjectPointer <XSpreadsheets2> l_spreadSheets = new UnoObjectPointer <XSpreadsheets2> ( (XSpreadsheets) l_originalUnoDocument.getAddress <XSpreadsheetDocument> ().getSheets ());
								Int32 l_numberOfSheets = l_spreadSheets.getAddress <XIndexAccess> ().getCount ();
								UnoObjectPointer <XNamed> l_spreadSheet = null;
								String l_targetFileNameAugmentation = null;
								Int32 l_lastPeriodPositionInTargetFileUrlBase = a_targetFileUrlBase.LastIndexOf (".");
								String l_augmentedTargetFileUrl = null;
								Boolean l_whetherSheetIsVisible = false;
								for (Int32 l_sheetIndex = GeneralConstantsConstantsGroup.c_iterationStartNumber; l_sheetIndex < l_numberOfSheets; l_sheetIndex ++) {
									l_spreadSheet = new UnoObjectPointer <XNamed> ( (Object) ( ( (Any) l_spreadSheets.getAddress <XIndexAccess> ().getByIndex (l_sheetIndex)).Value));
									if (a_hiddenSpreadSheetsAreWritten) {
									}
									else {
										l_whetherSheetIsVisible = (Boolean) (UnoDatumConverter.getObject (l_spreadSheet.getAddress <XPropertySet> ().getPropertyValue (UnoSpreadSheetPropertyNamesSet.c_whetherIsVisible_boolean)));
										if (!l_whetherSheetIsVisible) {
											continue;
										}
									}
									if (a_targetFileNamingRule == 0) {
										l_targetFileNameAugmentation = String.Format ("{0}", l_sheetIndex);
									}
									else {
										l_targetFileNameAugmentation = l_spreadSheet.getAddress ().getName (); 
									}
									if (l_lastPeriodPositionInTargetFileUrlBase >= GeneralConstantsConstantsGroup.c_iterationStartNumber) {
										l_augmentedTargetFileUrl = String.Format ("{0}_{1}{2}", a_targetFileUrlBase.Substring (GeneralConstantsConstantsGroup.c_iterationStartNumber, l_lastPeriodPositionInTargetFileUrlBase), l_targetFileNameAugmentation, a_targetFileUrlBase.Substring (l_lastPeriodPositionInTargetFileUrlBase));
									}
									else {
										l_augmentedTargetFileUrl = String.Format ("{0}_{1}", a_targetFileUrlBase, l_targetFileNameAugmentation);
									}
									if (l_sheetIndex != GeneralConstantsConstantsGroup.c_iterationStartNumber) {
										l_spreadSheets.getAddress ().moveByName (l_spreadSheet.getAddress ().getName (), (Int16) GeneralConstantsConstantsGroup.c_iterationStartNumber);
									}
									l_originalUnoDocument.getAddress <XStorable2> ().storeToURL (l_augmentedTargetFileUrl, a_unoDocumentStoringPropertiesArray);
								}
								l_hasSucceeded = true;
							}
						}
						catch (Exception l_exception) {
							throw l_exception;
						}
						l_originalUnoDocument.getAddress <XCloseable> ().close (false); 
					}
					else {
						Publisher.logErrorInformation (String.Format ("### The original file: '{0:s}' cannot be opened.", a_originalFileUrl));
					}
					return l_hasSucceeded;
				}
			}
		}
	}
}

